<?php

declare(strict_types=1);

namespace Drupal\idle_reauthenticate\Controller;

use Drupal\Core\Access\AccessResult;
use Drupal\Core\Access\AccessResultInterface;
use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Ajax\CloseModalDialogCommand;
use Drupal\Core\Ajax\OpenModalDialogCommand;
use Drupal\Core\Ajax\RedirectCommand;
use Drupal\Core\Ajax\SettingsCommand;
use Drupal\Core\Config\ImmutableConfig;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Password\PasswordInterface;
use Drupal\Core\TempStore\PrivateTempStore;
use Drupal\Core\Url;
use Drupal\idle_reauthenticate\Form\Authenticate;
use Drupal\user\Entity\User;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Request;

/**
 * Returns responses for Reauthenticate on idle browser routes.
 */
final class PingBack extends ControllerBase {

  public const string STATE_BLOCKED = 'blocked';
  public const string STATE_UNBLOCKED = 'unblocked';
  public const string STATE_UNKNOWN = 'unknown';

  /**
   * The ajax response.
   *
   * @var \Drupal\Core\Ajax\AjaxResponse
   */
  protected AjaxResponse $response;

  /**
   * The controller constructor.
   */
  public function __construct(
    protected readonly Request $request,
    protected readonly PrivateTempStore $store,
    protected readonly PasswordInterface $passwordChecker,
    protected readonly ImmutableConfig $config,
  ) {
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container): self {
    return new self(
      $container->get('request_stack')->getCurrentRequest(),
      $container->get('tempstore.private')->get('idle_reauthenticate'),
      $container->get('password'),
      $container->get('config.factory')->get('idle_reauthenticate.settings'),
    );
  }

  /**
   * Access controller for the ping back.
   *
   * @return \Drupal\Core\Access\AccessResultInterface
   *   Always allow access because we have to allow requests from e.g. inactive
   *   browser tabs when the focused tab logged out the current user. In such
   *   cases, the ping should be allowed but then also perform a redirect to the
   *   login page.
   */
  public function access(): AccessResultInterface {
    return AccessResult::allowed();
  }

  /**
   * Builds the response.
   */
  public function __invoke(): AjaxResponse {
    $this->response = new AjaxResponse();
    if (!$this->currentUser()->isAuthenticated()) {
      $this->reLogin();
      return $this->response;
    }
    $payload = $this->request->getPayload();
    if ($payload->has('form_id') && $payload->get('form_id') === 'idle_reauthenticate_authenticate') {
      $user = User::load($this->currentUser()->id());
      if ($payload->get('_triggering_element_name') === 'logout' || $user->isBlocked()) {
        $this->reLogin();
      }
      else {
        if ($this->passwordChecker->check($payload->get('password'), $user->getPassword())) {
          $this->unblockSession();
        }
      }
      return $this->response;
    }

    $lastSeen = $payload->get('lastSeen', 0);
    $currentTime = $payload->get('currentTime', 0);
    $currentState = $payload->get('currentState', FALSE);

    $dbState = $this->store->get('state') ?? self::STATE_UNKNOWN;
    $dbLastSeen = $this->store->get('lastSeen') ?? 0;
    if ($currentState === FALSE) {
      $this->store->set('lastSeen', $currentTime);
      if ($dbState === self::STATE_BLOCKED) {
        $this->blockSession();
      }
      else {
        $this->unblockSession(FALSE);
      }
    }
    else {
      if ($dbLastSeen > $lastSeen) {
        $lastSeen = $dbLastSeen;
      }
      else {
        $this->store->set('lastSeen', $lastSeen);
      }
      if ($currentState === self::STATE_UNBLOCKED
        && ($currentTime - $lastSeen) > $this->config->get('idle_period') * 1000
      ) {
        $this->blockSession();
      }
      elseif ($currentState === self::STATE_BLOCKED && $dbState === self::STATE_UNBLOCKED) {
        $this->unblockSession();
      }
    }
    return $this->response;
  }

  /**
   * Set the session blocked.
   */
  private function blockSession(): void {
    if (($this->store->get('state') ?? self::STATE_UNKNOWN) !== self::STATE_BLOCKED) {
      $this->store->set('state', self::STATE_BLOCKED);
    }
    $this->response->addCommand(new SettingsCommand(['idleReauthenticate' => ['currentState' => self::STATE_BLOCKED]], TRUE));
    $form = $this->formBuilder()->getForm(Authenticate::class);
    $this->response->addCommand(new OpenModalDialogCommand(
      $this->t('Re-authenticate %name', ['%name' => $this->currentUser()->getDisplayName()]),
      $form,
      [
        'classes' => [
          'ui-dialog' => 'idle-reauthenticate',
        ],
        'closeOnEscape' => FALSE,
      ]));
  }

  /**
   * Set the session unblocked.
   */
  private function unblockSession(bool $closeDialog = TRUE): void {
    if (($this->store->get('state') ?? self::STATE_UNKNOWN) !== self::STATE_UNBLOCKED) {
      $this->store->set('state', self::STATE_UNBLOCKED);
    }
    $this->response->addCommand(new SettingsCommand(['idleReauthenticate' => ['currentState' => self::STATE_UNBLOCKED]], TRUE));
    if ($closeDialog) {
      $this->response->addCommand(new CloseModalDialogCommand());
    }
  }

  /**
   * Logout the current user and redirect to the login page.
   */
  private function reLogin(): void {
    if ($this->currentUser()->isAuthenticated()) {
      user_logout();
      $this->store->delete('state');
    }
    $this->response->addCommand(new RedirectCommand(Url::fromRoute('user.login')->toString()));
  }

}
